<?php

/*
    Web Application Deployment System
    (c)2006-2008 Tim Jackson (tim@timj.co.uk)
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of version 3 of the GNU General Public License as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

define('WADF_VCREFTYPE_REV','rev');
define('WADF_VCREFTYPE_TAG','tag');
define('WADF_VCREFTYPE_BRANCH','branch');
define('WADF_VCREFTYPE_TRUNK','trunk');
define('WADF_VCREFTYPE_UNKNOWN','unknown');

/**
 * Main WADF class
 */
class Tools_WADF {

	var $_debug = 0;

	// the app reference
	var $appref;

	// macro definitions
	var $_macro_defs;
	
	// resolved macro values
	var $_macro_values = array();
	
	// generic class options
	var $_options = array(
		'master_config' => '@cfg_dir@/Tools_WADF/wadf.conf',
		'webserver_type' => 'httpd',
		'update_hosts' => false,
	);
	
	// macros which we don't manage to resolve during deployment
	var $_unresolved_macros = array();
	
	function Tools_WADF ($appref, $options=null, $cmdline_macros=null)
	{
		// Strip trailing slash from appref, if present
		if (substr($appref, -1, 1) == '/') {
			$appref = substr($appref, 0, strlen($appref)-1);
		}
		
		$valid_options = array('master_config');
		if (is_array($options)) {
			foreach ($options as $option => $value) {
				if (in_array($option, $valid_options)) {
					$this->_options[$option] = $value;
				} else {
					trigger_error("Invalid config option '$option' passed to constructor",E_USER_ERROR);
				}
			}
		}
		
		$this->_setInternalMacros();
		
		// set default options
		$this->_appendMacroDefs($this->_options);
		
		// set appref
		$this->appref = $appref;
		$options['appref'] = $appref;
		
		// set user-supplied options
		$this->_appendMacroDefs($options);

		if (!is_array($cmdline_macros)) {
			$cmdline_macros = array();
		}
		$this->processConfigs($cmdline_macros);
		
		$this->resolveAllMacros();
	}
	
	function _setInternalMacros()
	{
		if (isset($_ENV['HOSTNAME'])) {
			$macros['hostname'] = $_ENV['HOSTNAME'];
		} else {
			$macros['hostname'] = gethostbyaddr('127.0.0.1'); // FIXME this can't be the right way of doing it
		}
		$macros['cwd'] = getcwd();
		if (function_exists('posix_getuid')) {
			$details = posix_getpwuid(posix_getuid());
			$macros['user'] = $details['name'];
		} else {
			$macros['user'] = 'UNKNOWN';
		}
		if (isset($_ENV['HOME'])) {
			$macros['home'] = $_ENV['HOME'];
		}
		$this->_appendMacroDefs($macros);
	}
	
	// branch = [a branch], trunk, tags
	// rev = [rev no], HEAD
	function deploy($revtype, $rev_translated, $raw_rev=null, $db_deploy=false, $db_deploy_pass='')
	{
		// TODO: check if there is already a deployed instance of the app!
	
		$dir = $this->resolveMacro('deploy_path');
		$this->checkout($dir, $revtype, $rev_translated, $raw_raw);
		
		// The local config might be stored in the checkout, so check the local
		// config again
		$this->processLocalConfig();
		
		// Check for any config options that require user input
		$this->checkOptionsRequiringInput();
		
		$this->setupPEAR($dir);
		$this->processTemplatesInDir($dir);
		
		// TODO: handled unresolved macros? or just let UI do it?
		
		if ($db_deploy) {
			$this->deployDatabase($db_deploy_pass);
		}
		$this->deployVhost();
		$this->deployDNS();
		$this->deployScheduledJobs();
		$this->runKickstart();
	}
	
	function getUnresolvedMacros()
	{
		return $this->_unresolved_macros;
	}

	function unDeploy()
	{
		$this->unDeployDatabase();
		// TODO remove files, remove vhost, remove DNS
	}
	
	function _enumerateMultipleEntities()
	{
		$dir = $this->resolveMacro('deploy_path');
	
		$macros_in_templates = $this->extractMacrosFromDir($dir);
		// we need to enumerate the db and vhost numbers
		$in_use = array();
		foreach ($macros_in_templates as $macro) {
			if (preg_match('/^(db|vhost)(\d+)_/', $macro, $matches)) {
				// array like
				// in_use = array (
				//   db => array (1=>true, 2=>true, 3=>true)
				//   vhost => array (1=>true, 2=>true)
				$in_use[$matches[1]][$matches[2]] = true;
			}
		}
		// now turn the array into db => array(1,2,3) etc.
		foreach ($in_use as $type => $list_of_valid) {
			$in_use[$type] = array_keys($list_of_valid);
		}
		
		return $in_use;
	}
	
	function deployDatabase($rootpass='')
	{
		$in_use = $this->_enumerateMultipleEntities();
		$dir = $this->resolveMacro('deploy_path');
		
		// deploy database
		if (isset($in_use['db'])) {
			foreach ($in_use['db'] as $num) {
				$host = $this->resolveMacro("db${num}_host");
				$user = $this->resolveMacro("db${num}_user");
				$pass = $this->resolveMacro("db${num}_pass");
				$name = $this->resolveMacro("db${num}_name");
				$type = $this->resolveMacro("db${num}_type");
				$schema = $this->resolveMacro("db${num}_schema");
				if ($type != 'mysql') {
					die ("Unsupported database type '$type'");
				}
				// TODO check host, user, pass, name are all non-empty
				print "Setting up database $name on host $host...\n";
				$schema_path = "$dir/$schema";
				if (file_exists($schema_path)) {
					// FIXME abstract db-specific stuff
					// FIXME error handling!
					// FIXME quote strings!
					$db = @mysql_connect($host, 'root', $rootpass);
					if (!is_resource($db)) die("Could not connect to database: ".mysql_error()."\n");
					mysql_query("CREATE DATABASE $name", $db);
					mysql_query("GRANT ALL on $name.* to $user IDENTIFIED BY '$pass'", $db);
					mysql_close($db);
					$cmd = "mysql $name -h $host -u $user ".(($pass != null) ? "-p$pass" : '')." < $dir/$schema";
					//print "$cmd\n";
					exec($cmd);
				} else {
					die("ERROR: schema file $schema_path not found\n");
				}
			}
		} else {
			print "No database to deploy\n";
		}
		
	}
	
	function unDeployDatabase($rootpass='')
	{
		$in_use = $this->_enumerateMultipleEntities();
		if (isset($in_use['db'])) {
			foreach ($in_use['db'] as $num) {
				$host = $this->resolveMacro("db${num}_host");
				$name = $this->resolveMacro("db${num}_name");
				// TODO error handling
				$db = @mysql_connect($host, 'root', $rootpass);
				mysql_query("DROP DATABASE $name");
				mysql_close($db);
				// TODO remove DB users? But what if other site uses them?
			}
		}
	}
	
	function deployDNS()
	{
		// TODO other ways of deploying DNS
		$deploy_type = $this->resolveMacro('deploy_dns');

		// NB that the string "none" is transparently converted to the empty
		// string by the PHP INI file parser
		if ($deploy_type == '') {
			return;
		}

		$in_use = $this->_enumerateMultipleEntities();
		if (isset($in_use['vhost'])) {
			foreach ($in_use['vhost'] as $vhost_id) {
				$hosts[] = $this->resolveMacro("vhost${vhost_id}_name");
			}

			switch ($deploy_type) {
				case 'hosts':
					$this->addHostsToHostsFile($hosts);
					break;
				default:
					trigger_error("Unknown DNS deployment method '$deploy_type'",E_USER_ERROR);
					break;
			}
		}
	}
	
	function deployScheduledJobs()
	{
		$crontab_file = $this->resolveMacro('deploy_path') . '/crontab';
		if (file_exists($crontab_file)) {
			print "Deploying scheduled jobs from $crontab_file...\n";
			$instance = $this->resolveMacro('instance');
			$crontab_entries = trim(file_get_contents($crontab_file));
			exec('crontab -l 2>&1', $current_crontab, $ret);

			// The special markers we put at the start and end of the crontab			
			$wadf_deploy_begin = "# wadf-deployment-begin: $instance - DO NOT REMOVE THIS";
			$wadf_deploy_end = "# wadf-deployment-end: $instance - DO NOT REMOVE THIS";
			
			// The new entries we want to add
			$crontab_entries_new = "$wadf_deploy_begin\n$crontab_entries\n$wadf_deploy_end";

			if ($ret == 1 || (count($current_crontab) == 1 && preg_match('/^no crontab for/i', $current_crontab))) {
				// No current crontab
				print "\tDeploying new crontab\n";
				$current_crontab = '';
				$new_crontab = $crontab_entries_new . "\n";
			} else {
				$current_crontab = implode("\n", $current_crontab) . "\n";
				
				$regex = "/# wadf-deployment-begin: $instance.+# wadf-deployment-end: $instance - DO NOT REMOVE THIS/ms";
				if (preg_match($regex, $current_crontab)) {
					print "\tFound $instance in existing crontab, replacing\n";
					$new_crontab = preg_replace($regex, $crontab_entries_new, $current_crontab);
				} else {
					print "\tCould not find $instance in existing crontab\n";
					$new_crontab = $current_crontab . "\n" . $crontab_entries_new . "\n";
				}
			}
			$tmp_file = tempnam('/tmp', 'wadfcron');
			$fp = fopen($tmp_file, 'w');
			fputs($fp, $new_crontab);
			fclose($fp);
			exec("crontab $tmp_file");
			unlink($tmp_file);
		}
	}
	
	function addHostsToHostsFile($hosts)
	{
		$hosts_file = $this->resolveMacro('deploy_dns_hosts_file');
		$deploy_ip = $this->resolveMacro('deploy_dns_hosts_ip');
		$existing_hostnames = array();
		$file_contents = file($hosts_file);
		$line_num = 0;
		$localhost_line_num = 0;
		$base_host = ''; //the first host found
		foreach ($file_contents as $line) {
			if (substr($line, 0, strlen($deploy_ip)) == $deploy_ip) {
				if (preg_match_all('/[\t\s]+(\S+)/', $line, $matches)) {
					foreach ($matches[1] as $host) {
						$existing_hostnames[] = trim($host);
						if (!$base_host) $base_host = trim($host);
					}
				}
				$localhost_line_num = $line_num;
			}
			$line_num++;
		}
		foreach ($hosts as $host) {
			if (!in_array($host, $existing_hostnames)) {
				array_splice($file_contents, $localhost_line_num + 1, 0, "$deploy_ip\t$base_host\t$host\n");
			}
		}
		$fp = @fopen($hosts_file, 'w');
		if (is_resource($fp)) {
			fputs($fp, implode('', $file_contents));
			fclose($fp);
		} else {
			print "WARNING: Could not update list of local hostnames in hosts file '$hosts_file'. Check permissions.\n";
		}
	}
	
	function deployVhost($dir=null)
	{
		if ($dir === null) {
			$dir = $this->resolveMacro('deploy_path');
		}

		// deploy vhost config
		$source_file = $dir.'/'.$this->resolveMacro('vhost_config_template');
		$dest_file = $this->resolveMacro('vhost_config_path').'/'.$this->resolveMacro('instance').'.conf';
		//print "Copying $source_file to $dest_file\n";
		
		$config = file_get_contents($source_file);
		
		// If we are using mod_php, insert the PHP config options
		if ($this->resolveMacro('php_type') == 'mod_php') {
			$php_ini = $dir . '/' . $this->resolveMacro('php_config_location');
			$directives = $this->_convertPhpIniToApacheModPhp($php_ini);
			
			// Stick the PHP directives at the end of the vhost
			$php_directives = implode("\n\t", $directives);
			$config = str_replace("</VirtualHost>", "\n\t# PHP directives processed from $php_ini\n\t" . $php_directives . "\n" . '</VirtualHost>', $config);
			
			// Process local PHP override ini file, if it exists
			$php_ini_local = $this->resolveMacro('php_config_location_extra');
			if ($php_ini_local != '@php_config_location_extra@') {
				$directives_local = $this->_convertPhpIniToApacheModPhp($php_ini_local);
				if (count($directives_local) > 0) {
					$php_directives = implode("\n\t", $directives_local);
					$config = str_replace("</VirtualHost>", "\n\t# PHP directives processed from $php_ini_local\n\t" . $php_directives . "\n" . '</VirtualHost>', $config);
				}
			}
		} else {
			$this->_deployPHPConfig($dir);
		}
		
		$deploy_version = $this->resolveMacro('deploy_version');
		if ($deploy_version == '@deploy_version@') $deploy_version = 'unknown';
		$config = "# wadf-working-copy: $dir\n# wadf-deploy-version: $deploy_version\n$config";
		$fp = fopen($dest_file, 'w');
		if (is_resource($fp)) {
			fputs($fp, $config);
			fclose($fp);
		} else {
			print "Could not open vhost config destination file '$dest_file'\n";
		}
		
		// restart apache
		$restart_cmd = $this->resolveMacro('webserver_restart_cmd');
		print "Restarting webserver...\n";
		if ($restart_cmd != '@webserver_restart_cmd@') {
			passthru($restart_cmd);
		}
	}
	
	/**
	 * @param string $file Filename to read. If file does not exist, no error will be thrown.
	 * @return array Array of Apache mod_php config rules
	 */
	function _convertPhpIniToApacheModPhp($file)
	{
		$directives_out = array();
		if (file_exists($file)) {
			print "Processing PHP config file $file...\n";
			$php_ini_directives = file($file);
			$directives_out = array();
			foreach ($php_ini_directives as $directive) {
				if (preg_match('/^\s*;(.+)$/', $directive, $matches)) {
					// Comment; convert php.ini-style semicolons to hash signs
					$directives_out[] = '#' . $matches[1];
				} else if (preg_match('/^([a-z0-9_]+)\s*=\s*(.+)$/', $directive, $matches)) {
					$directive_name = trim($matches[1]);
					$directive_value = trim($matches[2]);
					
					// See if it looks like a flag or a value
					if (in_array(strtolower($directive_value), array('0','1','on','off'))) {
						$directives_out[] = "php_admin_flag $directive_name $directive_value";
					} else {
						$directives_out[] = "php_admin_value $directive_name $directive_value";
					}
				} else {
					print "WARNING: Could not parse PHP ini config line from $file:\n\t$directive\n";
				}
			}
		}
		return $directives_out;
	}
	
	// Not needed for mod_php; it is done in deployVhost();
	function _deployPHPConfig($dir)
	{
		$php_type = $this->resolveMacro('php_type');
		if ($php_type == 'mod_php') {
			return; // do nothing
		} else if (preg_match('/^cgi:(.+)$/', $php_type, $matches)) {
			$php_ini_dest = trim($matches[1]);
			$php_ini_source = $dir . '/' . $this->resolveMacro('php_config_location');
			if (file_exists($php_ini_source)) {
				print "Deploying PHP config file $php_ini...\n";
				copy($php_ini_source, $php_ini_dest);
			}
		} else {
			print "ERROR: Unknown PHP type '$php_type'\n";
		}
	}
	
	// FIXME abstract
	/**
	 * Check out something from version control system
	 * @param string $destdir         The destination to check a working copy out to
	 * @param string $revtype         One of the WADF_VCREFTYPE_* constants
	 * @param string $rev_translated  The "translated" revision which makes sense according to $revtype. For tags, this is the tag. For branches, the branch name. For trunk, pass NULL.
	 * @param string $raw_rev         Raw changeset number. Not valid for tags. Defaults to HEAD.
	 */
	function checkout ($destdir, $revtype, $rev_translated, $raw_rev=null)
	{
		// MODULE is appref
		// BASE URL is vc_base
		
		switch ($revtype) {
			case WADF_VCREFTYPE_TRUNK:
				$svn_path = 'trunk';
				break;
			case WADF_VCREFTYPE_BRANCH:
				$svn_path = "branches/$rev_translated";
				break;
			case WADF_VCREFTYPE_TAG:
				$svn_path = "tags/$rev_translated";
				if ($raw_rev !== null) {
					trigger_error("Don't pass a raw revision when checking out a tag", E_USER_ERROR);
				}
				break;
		}
		
		if ($raw_rev !== null) {
			$options = "-r $raw_rev";
		}
		
		$vc_base = $this->resolveMacro('vc_base');
		
		// See if we already have a checked-out copy
		// We could just do a file_exists() on .wadf-instance, but actually
		// checking for SVN metadata is safer
		$vc_info = $this->_readVCInfo($destdir);
		
		$svn_action = 'checkout';
		if (is_array($vc_info)) {
			$svn_action = 'switch';
		}
		
		if ($svn_action == 'checkout') {
			exec("rm -rf $destdir");
		} else {
			// FIXME: we should do a wadf-clean before doing an svn sw, to stop
			// errors about "file already exists" in some cases
		}
		
		$cmd = "svn $options $svn_action $vc_base/$this->appref/$svn_path $destdir";
		print "Running $cmd\n";
		exec($cmd, $output, $ret);
		print implode("\n",$output)."\n";
		if ($ret != 0) {
			print "Error checking out\n";
			exit(5);
		}
		
		if (file_exists("$destdir/.wadf-instance")) {
			$instance_file = file("$destdir/.wadf-instance");
			$instance_file[0] = $this->appref;
			if ($fp = @fopen("$destdir/.wadf-instance",'w')) {
				fputs($fp, implode('', $instance_file));
				fclose($fp);
			}
		} else {
			if ($fp = @fopen("$destdir/.wadf-instance",'w')) {
				fputs($fp, $this->appref);
				fclose($fp);
			}
		}		
		
		$this->setVCVersionMacro($destdir);
		
	}
	
	/**
	 * Get a unique identifying version string and set it as a macro
	 * @return string The version control identifier
	 */
	function setVCVersionMacro($dir)
	{
		$live_vc_info = $this->_readVCInfo($dir);
		$vc_id = $this->_getVCIdentifier($live_vc_info['rev_type'], $live_vc_info['rev_translated'], $live_vc_info['rev_raw']);
		$this->_appendMacroDefs(array('deploy_version' => $vc_id));
		return $vc_id;
	}
	
	/**
	 * Read version control info from deployed copy
	 * @return array
	 */
	function _readVCInfo($dir)
	{
		// FIXME: support more than just SVN
		$cmd = "svn info $dir 2>/dev/null";
		exec($cmd, $output);
		
		$info['url'] = '';
		$info['rev_raw'] = '';
		$info['rev_translated'] = '';
		$info['rev_type'] = '';
		
		foreach ($output as $line)
		{
			if (preg_match('/^URL: (.+)$/', $line, $matches)) {
				$info['url'] = trim($matches[1]);
			} else if (preg_match('/^Last Changed Rev: (\d+)$/', $line, $matches)) {
				$info['rev_raw'] = trim($matches[1]);
			}
		}
		
		if (empty($info['url']) || empty($info['rev_raw'])) {
			//print "Could not read VC info from $dir\n";
			return false;
		}
		
		if (preg_match('|/trunk$|', $info['url'])) {
			$info['rev_type'] = WADF_VCREFTYPE_TRUNK;
		} else if (preg_match('|/tags/([^/]+)$|', $info['url'], $matches)) {
			$info['rev_type'] = WADF_VCREFTYPE_TAG;
			$info['rev_translated'] = $matches[1];
		} else if (preg_match('|/branches/([^/]+)$|', $info['url'], $matches)) {
			$info['rev_type'] = WADF_VCREFTYPE_BRANCH;
			$info['rev_translated'] = $matches[1];
		} else {
			print "Could not work out VC rev type from URL ". $info['url'] ."\n";
			$info['rev_type'] = WADF_VCREFTYPE_UNKNOWN;
			$info['rev_translated'] = 'unknown';
		}
		
		return $info;
	}
	
	function _getVCIdentifier($revtype, $rev_translated, $rev_raw)
	{
		switch ($revtype) {
			case WADF_VCREFTYPE_TRUNK:
				$output = 'DEVTR' . ':' . $rev_raw;
				break;
			case WADF_VCREFTYPE_BRANCH:
				$output = 'DEVBR' . '/' . $rev_translated . ':' . $rev_raw;
				break;
			case WADF_VCREFTYPE_TAG:
				$output = $rev_translated;
				break;
			case WADF_VCREFTYPE_UNKNOWN:
				$output = 'unknown';
				break;
		}
		return $output;
	}

	function processVCVersionString($string, $rev='HEAD')
	{	
		$out['rev_type'] = WADF_VCREFTYPE_TRUNK;
		$out['rev_translated'] = null;
		
		if (preg_match('#^(trunk|tag|branch)(/(.+))?$#', $string, $matches)) {
			if (isset($matches[3])) {
				$out['rev_translated'] = trim($matches[3]);
			}
			
			switch ($matches[1]) {
				case 'trunk':
					if (isset($matches[2])) {
						print "trunk does not take a parameter\n";
						exit(5);
					}
					break;
				case 'tag':
					$out['rev_type'] = WADF_VCREFTYPE_TAG;
					if ($rev != 'HEAD') {
						return "Don't pass a revision number when deploying from a tag\n";
					}
					
					if (empty($out['rev_translated'])) {
						return "Missing tag number\n";
					}
					break;
				case 'branch':
					$out['rev_type'] = WADF_VCREFTYPE_BRANCH;
					if (empty($out['rev_translated'])) {
						return "Missing branch name\n";
					}
					break;
			}
		} else {
			return "Could not parse revision type '$string'\n";
		}
		return $out;
	}
	
	function setupPEAR($dir)
	{
		// See if we actually need PEAR for this deployment
		if (!file_exists("$dir/package.xml")) {
			return;
		}
		
		$pear_path = $this->resolveMacro('dep_pear_deploy_path');
		$standalone_pear = false;
		// if empty, we want to use already-installed PEAR
		if (!empty($pear_path)) {
			$pear_config_file = $this->resolveMacro('dep_pear_config_file');
			if (empty($pear_config_file)) {
				die("No dep_pear_config_file option specified");
			}
			$standalone_pear = true;
			if (!file_exists($pear_config_file)) {
				$cmd = "pear -v -c $pear_config_file ".
					"-d php_dir=$pear_path/php ".
					"-d data_dir=$pear_path/data ".
					"-d ext_dir=$pear_path/ext ".
					"-d doc_dir=$pear_path/docs ".
					"-d test_dir=$pear_path/tests ".
					"-d cache_dir=$pear_path/cache ".
					"-d download_dir=$pear_path/downloads ".
					"-d temp_dir=$pear_path/temp ".
					"-d bin_dir=$pear_path ".
					"-s";
				exec($cmd);
					
				print "Installing PEAR Core into $pear_path (config file is $pear_config_file)...\n";
				passthru("pear -c $pear_config_file install --onlyreqdeps pear.php.net/PEAR");
			}
			
			$base_channel = $this->resolveMacro('dep_pear_base_channel');
			if (!empty($base_channel)) {
				if (preg_match('/^(.+):(.+)@(.+)$/',$base_channel, $matches)) {
					$username = $matches[1];
					$password = $matches[2];
					$channel = $matches[3];
				} else {
					$channel = $base_channel;
				}
				if (empty($channel)) trigger_error("Channel empty", E_USER_ERROR);
				print "Discovering channel $channel...\n";
				
				// FIXME check if channel is already discovered and skip next step if so
				
				// With PEAR 1.6.x+ the base_channel can include username and
				// password in the format username:password@channel
				$this->_runPEAR("channel-discover $base_channel");
				if ($username) {
					print "Logged in as user $username\n";
				}
			}
		} else {
			exec($this->_getPEARCmd() . ' config-get bin_dir', $output);
			$pear_path = $output[0];
			echo "Using existing PEAR installation in $pear_path\n";
		}
		
		$this->_installPEARBaseRoles();
		$this->_configurePEARBaseRoles($pear_path, $dir, $standalone_pear);
		
		print "Processing package.xml and installing PEAR dependencies...\n";
		$application_dir = '';
		if (!$standalone_pear) {
			// Deploy to the same directory. Is !$standalone_pear really the
			// right criteria to use here?
			$application_dir = "-d application_dir=$dir";
		}
		
		$this->_runPEAR("$application_dir $pear_stability upgrade --onlyreqdeps -f $dir/package.xml");
		
		$this->_setPearMacros();
		
	}
	
	function _installPEARBaseRoles()
	{
	
		$base_roles = $this->resolveMacro('dep_pear_base_roles');
		
		if (!empty($base_roles)) {
			// The command line binary to run
			$pear = $this->_getPEARCmd();
			
			$roles = explode(' ', $base_roles);
			foreach ($roles as $package) {
				exec("$pear list $package", $out, $ret);
				if ($ret != 0) {
					print "Installing base role package $package...\n";
					$this->_runPEAR("install --soft --onlyreqdeps $base_roles");
				}
			}
		}
	}
	
	function _configurePEARBaseRoles($pear_path, $dir, $standalone_pear=false)
	{
		if (empty($pear_path)) trigger_error("Empty pear_path passed to Tools_WADF::_configurePEARBaseRoles()",E_USER_ERROR);
	
		$vars = array(
			'applicationcore_dir' => $pear_path . '/applications'
		);
		
		if ($standalone_pear) {
			$vars['application_dir'] = $dir;
		}
		
		$base_channel = $this->_getPEARBaseChannelName();
		
		// The command line binary to run
		$pear = $this->_getPEARCmd();
		
		// Set the variables if needed
		foreach ($vars as $var => $value) {
			unset($out);
			unset($ret);
			exec("$pear config-get -c $base_channel $var", $out, $ret);
			if (empty($out[0]) || $standalone_pear) {
				print "Setting $var = $value\n";
				if (!empty($base_channel)) {
					$this->_runPEAR("config-set -c $base_channel $var \"$value\"");
				} else {
					$this->_runPEAR("config-set $var \"$value\"");
				}
			} else {
				print "Base role config variable $var is already set ($out[0]) - not changing\n"; 
			}
		}
	}
	
	/**
	 * Returns base channel name *without* any optional username/pass
	 */
	function _getPEARBaseChannelName()
	{
		$channel = null;
		$base_channel = $this->resolveMacro('dep_pear_base_channel');
		if (!empty($base_channel)) {
			if (preg_match('/^(.+):(.+)@(.+)$/',$base_channel, $matches)) {
				$channel = $matches[3];
			} else {
				$channel = $base_channel;
			}
			if (empty($channel)) trigger_error("Channel empty", E_USER_ERROR);
		}
		return $channel;
	}
	
	function _runPEAR($cmd)
	{
		$pear_stability = $this->resolveMacro('dep_pear_preferred_stability');
		if ($pear_stability) {
			$pear_stability = "-d preferred_state=$pear_stability";
		}
		
		$runcmd = $this->_getPEARCmd() . " $pear_stability " . $cmd;
		if ($this->_debug) print "Running: $runcmd\n";
		passthru($runcmd);
		return $ret;
	}
	
	function _getPEARCmd()
	{
		$pear_path = $this->resolveMacro('dep_pear_deploy_path');
		if (empty($pear_path)) {
			exec('pear config-get bin_dir', $output, $ret);
			$pear_path = $output[0];
			$runcmd = "$pear_path/pear";
		} else {
			$pear_config_file = $this->resolveMacro('dep_pear_config_file');
			$runcmd = "$pear_path/pear -c $pear_config_file";
		}
		return $runcmd;
	}
	
	/**
	 * Add a selection of PEAR config options to the macro list
	 */
	function _setPearMacros()
	{
		$macro_defs = array();
	
		// Technically we should have an option here to decide whether we are 
		// going to use PEAR's value for application_dir (i.e. the 
		// directory that PEAR installs the actual end client app into), or use
		// the actual working copy as the "real" live application. However,
		// for the time being we force it to be deploy_path (the working copy
		// checked out from version control) for simplicity.
		$macro_defs['application_dir'] = $this->resolveMacro('deploy_path');
	
		$pear_macros = array('applicationcore_dir', 'php_dir');
		
		$base_channel = $this->_getPEARBaseChannelName();
		
		foreach ($pear_macros as $macro) {
			$pear = $this->_getPEARCmd();
			unset($out);
			exec("$pear config-get -c $base_channel $macro", $out);
			if (!empty($out[0])) {
				$macro_defs[$macro] = trim($out[0]);
			}
		}
		$this->_appendMacroDefs($macro_defs);
	}
	
	function processTemplatesInDir($dir)
	{
		$files = $this->_listTemplateFiles($dir);
		foreach ($files as $file) {
			$content = file_get_contents($file);
			$content = $this->resolveString($content, null, "file:$file");
			$output_file = preg_replace('/^(.+)\.template$/','\1',$file);
			$fout_fh = fopen($output_file, 'w+');
			fputs($fout_fh, $content);
			fclose($fout_fh);
		}
	}
	
	// go through all template files in $dir and find out which macros are used in them
	// returns an array of unique macros
	function extractMacrosFromDir($dir)
	{
		$macros = array();
		$files = $this->_listTemplateFiles($dir);
		foreach ($files as $file) {
			if ($this->_debug) $this->_debugOutput("extracting macros from file '$file'");
			$contents = file_get_contents($file);
			$new_macros = $this->extractMacrosFromString($contents);
			if ($this->_debug) $this->_debugOutput("macros in file '$file' = ".implode(',',$new_macros));
			$macros = array_merge($macros,$new_macros);
			unset($new_macros);
		}
		return $macros;
	}
	
	function _listTemplateFiles($dir)
	{
		exec("find $dir -name '*.template'", $output, $ret);
		return $output;
	}
	
	// takes a string like '@underscore:foo@@bar@' and returns array('underscore:foo','bar');
	// if keep_modifiers = false then returns array('foo','bar');
	function extractMacrosFromString($string, $keep_modifiers = true)
	{
		preg_match_all('/@(([a-z]+:)?([a-z0-9_]{3,30}))@/', $string, $matches, PREG_PATTERN_ORDER);
		
		$macros = array_unique($matches[1]);
		if ($keep_modifiers == false) {
			foreach ($macros as $macro) {
				$newmacros[] = $this->getBaseMacroName($macro);
			}
			$macros = $newmacros;
		}
		return $macros;
	}
	
	function runKickstart()
	{
		$kickstart_script = $this->resolveMacro('kickstart_script');
		if (!empty($kickstart_script)) {
			// Nasty hack to make sure the kickstart script is executable.
			// Really it should be installed with role=script and PEAR would
			// handle it, but then it would end up being installed in the wrong
			// place
			// We should really read the existing mode and just set the executable
			// bit, but that is really hard to do with PHP because chmod() doesn't
			// like a variable as the second parameter very much.
			chmod($kickstart_script, 0750);
			
			print "\nRunning kickstart script $kickstart_script...\n";
			
			// Run the script
			passthru($kickstart_script);
		}
	}

	// $options is an array of macros e.g. passed from command line
	function processConfigs($options=null)
	{
		/* get:
		- globals
		- per-user
		- profile
		- app-specific (local_config)
		- instance-specific (.wadf-instance)
		- command line
		in that order. lump all together linearly in above order
		then process line-by-line, with later ones taking precedence.
		*/
		
		if ($options === null) $options = array();
		
		// Read global config
		$this->readConfigFile($this->_options['master_config']);
		
		// Read per-user config
		if (isset($_ENV['HOME'])) {
			$wadfrc = $_ENV['HOME'].'/.wadf/config';
			if (file_exists($wadfrc)) {
				$this->readConfigFile($wadfrc);
			}
		}
		
		// Read site-specific file
		// This is likely to reference deploy_path, so check that hasn't been
		// overridden - bit of a hack
		if (isset($options['deploy_path'])) {
			$this->_appendMacroDefs(array('deploy_path' => $options['deploy_path']));
		}
		$this->processLocalConfig();
		
		// Read site instance file
		$deploy_path = $this->resolveMacro('deploy_path');
		if (file_exists("$deploy_path/.wadf-instance")) {
			$instance_file = file("$deploy_path/.wadf-instance");
			unset($instance_file[0]);	// first line is the appref
			$new_macros = array();
			foreach ($instance_file as $line) {
				if (preg_match("/^(\S+)\s*=\s*(.*)$/", $line, $matches)) {
					$new_macros[$matches[1]] = $matches[2];
				}
			}
			if (count($new_macros) > 0) {
				$this->_appendMacroDefs($new_macros);
			}
		}
		
		// Append options from command line
		$this->_appendMacroDefs($options);
	}
	
	/**
	 * Process the config file from the local_config variable, if it exists
	 * @return void
	 */
	function processLocalConfig()
	{
		$app_conf_file = $this->resolveMacro('local_config');
		print "Looking for $app_conf_file...\n";
		if (file_exists($app_conf_file)) {
			print "\tFound. Loading configs from $app_conf_file\n";
			$this->readConfigFile($app_conf_file);
		} else {
			print "\tNot found\n";
		}
	}
	
	/**
	 * Check through the entire list of options for any which require user input
	 */
	function checkOptionsRequiringInput()
	{
		$input_values = array();
		foreach ($this->_macro_values as $key => $value) {
			// TODO: make %% into a const
			if ($value == '%%') {
				print "Please enter the value for the config option '$key': ";
				$input_values[$key] = fgets(STDIN);
			}
		}
		
		if (count($input_values) > 0) {
			$deploy_path = $this->resolveMacro('deploy_path');
			$instance_file = $deploy_path . '/.wadf-instance';
			$fp = fopen($instance_file, 'a');
			if (!is_resource($fp)) {
				die("Could not write to local instance file $instance_file");
			}
			foreach ($input_values as $key => $value) {
				fputs($fp, "$key = $value");
			}
			fclose($fp);
			$this->_appendMacroDefs($input_values);
		}
	}
	
	/**
	 * Go through vhost_config_path and analyse each vhost config, returning
	 * a set of information. Currently only works for Apache-format vhosts.
	 * @param string Vhost path (should be vhost_config_path, but that is not read automatically so this function can be called statically)
	 * @return array Array of vhosts, in the format:
	 *               ('appref' => array(), 'appref2' => array())
	 *               where the array for each appref has the following params:
	 *               working_dir
	 *               deploy_version
	 *               vhosts (array of hostname, interface, description[optional])
	 */
	function getDeployedVhosts($vhost_dir)
	{
		$dir = dir($vhost_dir);
		$vhosts = array();
		while (false !== ($entry = $dir->read())) {
			if (substr($entry, -5) == '.conf') {
				$vhost_details = array(
					'working_dir' => 'unknown',
					'deploy_version' => 'unknown',
					'vhosts' => array()
				);
			
				$string = @file_get_contents("$vhost_dir/$entry");
				if (!empty($string)) {
					$appname = str_replace('.conf','', $entry);
					
					// The dir that the app was deployed from
					if (preg_match('/^\s*#\s*wadf-working-copy:(.+)/m', $string, $matches)) {
						$vhost_details['working_dir'] = trim($matches[1]);
					}
					
					// Version control version
					if (preg_match('/^\s*#\s*wadf-deploy-version:(.+)/m', $string, $matches)) {
						$vhost_details['deploy_version'] = trim($matches[1]);
					}
					
					if (preg_match_all("/^<VirtualHost (.+)>(.+ServerName (.+)\s.+)<\/VirtualHost>/Uism", $string, $matches, PREG_SET_ORDER)) {
						foreach ($matches as $match_set) {
						
							$specific_vhost_details = array(
								'hostname' => $match_set[3],
								'interface' => $match_set[1]
							);
							// Vhost description
							if (preg_match('/^\s*#\s*desc:(.+)/m', $match_set[2], $matches_desc)) {
								$specific_vhost_details['description'] = trim($matches_desc[1]);
							}
							
							$vhost_details['vhosts'][] = $specific_vhost_details;
						}
						$vhosts[$appname] = $vhost_details;
					}
				}
			}
		}
		return $vhosts;
	}
	
	function _appendMacroDefsTop($array)
	{
		$copy = $this->_macro_defs;
		$this->_macro_defs = array();
		foreach ($array as $key => $value) {
			$this->_macro_defs[] = new Tools_WADF_MacroDef($key,$value);
		}
		foreach ($copy as $value) {
			$this->_macro_defs[] = $value;
		}
		$this->_rebuildMacroValues();
	}
	
	function _appendMacroDefs($array)
	{
		if (!is_array($array)) return;
		
		foreach ($array as $key => $value) {
			$this->_macro_defs[] = new Tools_WADF_MacroDef($key,$value);
		}
		$this->_rebuildMacroValues();
	}
	
	function _rebuildMacroValues()
	{
		if (count($this->_macro_values) > 0) {
			$this->resolveAllMacros();
		}
	}
	
	function readConfigFile($file)
	{
		$config = parse_ini_file($file, TRUE);
		if (isset($config['globals'])) {
			print "Setting global configuration options from file $file\n";
			$this->_appendMacroDefs($config['globals']);
		}
		// We need to do this to ensure "profile" is set.
		// Possible optimisation: treat "profile" specially, as an unexpandable
		// macro, and create a method something like "getLatestMacro" which gets
		// the most recent definition of a given "macro", and use the literal
		// contents of that.
		$this->resolveAllMacros();
		$profile = $this->resolveMacro('profile');
		if ($profile == '@profile@') {
			print "Cannot resolve global profile when processing $file\n";
			$profile = null;
		}
		
		// if no profile name passed, see if one was defined in the 'globals'
		// section of the config for the file that we're currently reading
		if ($profile == null) {
			if (isset($config['globals']['profile'])) {
				$profile = $config['globals']['profile'];
				print "File $file: Using profile='$profile' from 'globals' section of config\n";//TODO
			} else {
				print "File $file: ignoring as no global profile set\n";
			}
		} else {
			print "File $file: using profile section '$profile'\n";
		}
		
		if (isset($config[$profile])) {
			$this->_appendMacroDefs($config[$profile]);
		}
	}
	
	function _debugOutput($string)
	{
		print "$string\n";
	}
	
	// $string is a string which may contain all sorts of stuff including macros
	function resolveString ($string, $macro_values_stack = null, $context=null) {
		if ($macro_values_stack === null) {
		if ($this->_debug > 1) $this->_debugOutput("using default stack to resolve string");
			$macro_values_stack = &$this->_macro_values;
		}
		return $this->_resolveString($string, $macro_values_stack, $context);
	}
	
	// $context is an optional context for recording errors
	function _resolveString($string, &$macro_values_stack, $context=null)
	{
		if ($this->_debug > 1) $this->_debugOutput("resolving string '$string'");

		// take a copy for debugging
		$origstring = $string;
		
		$oldstring = '';
		while ($string != $oldstring) {
			$oldstring = $string;
			$macros = $this->extractMacrosFromString($string);
			if ($this->_debug > 1) $this->_debugOutput("macros in string '$string' = ".implode(',',$macros));
			foreach ($macros as $macro) {
				$macro_contents = $this->_resolveMacro($macro, $macro_values_stack);
				if ("@$macro@" == $macro_contents) {
					// Didn't manage to resolve macro
					$hash = md5($context . $macro); // prevent duplicates
					$this->_unresolved_macros[$hash] = array('macro' => $macro, 'context' => $context);
				} else {
					$string = str_replace("@$macro@", $macro_contents, $string);
				}
			}
		}
		if ($this->_debug > 1) $this->_debugOutput("resolved string '$origstring' to '$string'");
		return $string;
	}
	
	function resolveAllMacros()
	{
		$changed = true;
		
		foreach ($this->_macro_defs as $macro) {
			$this->_macro_values[$macro->name] = $macro->value;
		}
		
		while ($changed) {
			if ($this->_debug > 1) $this->_debugOutput("resolveAllMacros: iterating");
			$changed = false;
			foreach ($this->_macro_values as $macro_name => $macro_value) {
				if ($this->_debug > 1) $this->_debugOutput("resolveAllMacros: resolving macro '$macro->name' ('$macro->value')");
				$this->_macro_values[$macro_name] = $this->resolveString($macro_value);
			
				// If macro value has changed, we will need to iterate again
				if ($this->_macro_values[$macro_name] != $macro_value) {
					$changed = true;
				}
			
				if ($this->_debug > 1) $this->_debugOutput("resolveAllMacros: resolved $macro->name to '".$this->_macro_values[$macro->name]."'");
			}
		}
		return $this->_macro_values;
	}
	
	// $macro is a SINGLE macro (not a string containing macros) something like "foo" or "hyphen:foo"
	// note "foo" not "@foo@"
	// NB no recursive replace is required here because by this stage ALL MACROS should already have been expanded
	
	function resolveMacro ($macro, $macro_values_stack = null) {
		if ($macro_values_stack === null) {
			$macro_values_stack = &$this->_macro_values;
		}
		return $this->_resolveMacro($macro, $macro_values_stack);
	}
	
	function _resolveMacro ($macro, &$macro_values_stack) {
		if ($this->_debug > 1) $this->_debugOutput("resolve single macro '$macro'");
	
		// make a copy
		$origmacro = $macro;
		
		$output = null;
	
		# separate the modifier if macro is "foo:bar" 
		if (preg_match('/([a-z]+):(.+)/',$macro,$matches)) {
			$modifier = $matches[1];
			$macro = $matches[2];
		}
		
		if (isset($macro_values_stack[$macro])) {
			if ($this->_debug) $this->_debugOutput("found macro '$macro'");
			$output = $macro_values_stack[$macro];
		} else {
			$macro_fallbacks = array(
				'vhost(\d+)_name' => 'vhost_name',
				'vhost(\d+)_interface' => 'vhost_interface',
				'vhost(\d+)_config_template' => 'vhost_config_template',
				'db(\d+)_type' => 'db_type',
				'db(\d+)_name' => 'db_name',
				'db(\d+)_host' => 'db_host',
				'db(\d+)_user' => 'db_user',
				'db(\d+)_pass' => 'db_pass',
				'db(\d+)_schema' => 'db_schema',
			 );
			$fallback_match = false;
			foreach ($macro_fallbacks as $fallback_regex => $fallback_macro) {
				if (!$fallback_match) {
					if ($this->_debug) $this->_debugOutput("trying fallback '$fallback_regex'");
					if (preg_match("/^$fallback_regex\$/", $macro, $matches)) {
						$enum_value = $matches[1];
						if ($this->_debug) $this->_debugOutput("matched fallback '$fallback_regex' - numbered '$enum_value'");
						if ($enum_value == 1) $enum_value = '';
						
						// Now fill in the special enumerated macro with the relevant number
						if (substr($fallback_macro,0,2) == 'db') {
							$special_macro = 'db_number';
						} else {
							$special_macro = 'vhost_number';
						}
						
						// Fork off a new macro values stack
						
						if ($this->_debug) $this->_debugOutput("Forking new macro values stack with $special_macro = $enum_value");
						$macro_values_special = $macro_values_stack;
						
						$macro_values_special[$special_macro] = $enum_value;
						
						$output = $this->_resolveString("@$fallback_macro@", $macro_values_special);
						$fallback_match = true;
					}
				}
			}
		}
		
		$output_unprocessed = $output;
		// if output is still null, output is just macro name
		if ($output === null) {
			$output = "@$origmacro@";
			$output_unprocessed = $output;
		} else {
			// post-process output with modifier if relevant
			if (isset($modifier)) {
				switch($modifier) {
					case 'hyphen':
						$sub = '-';
						break;
					case 'slash':
						$sub = '/';
						break;
					case 'underscore':
						$sub = '_';
						break;
					default:
						trigger_error("unknown modifier $modifier",E_USER_ERROR);
				}
				if (isset($sub)) {
					$output = str_replace($this->_resolveMacro('appref_separator', $macro_values_stack), $sub, $output);
				}
			}
		}
		
		if ($this->_debug) $this->_debugOutput("resolved single macro '$origmacro' to '$output'");
		
		$macro_values_stack[$macro] = $output_unprocessed;
		return $output;
	}
	
	
	/**
	 * Get macro name without modifiers
	 *
	 * @param string $macro_name Macro name, possibly including modifiers
	 */	
	function getBaseMacroName($macro_name)
	{
		preg_match('/([a-z]+):(.+)/', $macro_name, $matches);
		return $matches[2];
	}
	
}

class Tools_WADF_MacroDef
{
	var $name;
	var $value;
	
	function Tools_WADF_MacroDef($name, $value)
	{
		$this->name = $name;
		$this->value = $value;
	}
}

?>
